#include <windows.h>
#include "pixelfun/pixelfun.h"
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include "Vector.h"

#define rnd(n) (rand()%(n+1))

//#define WINWIDTH 1280
//#define WINHEIGHT 960
#define WINWIDTH 640
#define WINHEIGHT 480
//#define WINWIDTH 320
//#define WINHEIGHT 240

#define BUFFERWIDTH (WINWIDTH+16)
#define BUFFERHEIGHT (WINHEIGHT+16)


Vector Normalize(Vector a)
{
	a.Normalize();
	return a;
}

DWORD* screen;

void ClearScreen(DWORD color)
{
	for(int i=0;i<WINWIDTH*WINHEIGHT;i++)
		screen[i]=color;
}

void DrawBar(int sx, int sy, int w, int h, DWORD color)
{
	if(sx<0) { w+=sx; sx=0; }
	if(sy<0) { h+=sy; sy=0; }
	if(sx+w>WINWIDTH) { w-=sx-WINWIDTH; }
	if(sy+h>WINHEIGHT) { h-=sy-WINHEIGHT; }
	for(int y=sy;y<sy+h;y++)
	{
		DWORD* sp=screen+y*WINWIDTH+sx;
		for(int x=sx;x<sx+w;x++)
			*sp++=color;
	}
}


struct FPixel
{
	int red, green, blue;
};

DWORD PixelDWORD(FPixel& p)
{
	return ((p.red<<8)&0xFF0000)|((p.green<<0)&0xFF00)|(p.blue>>8);
}

FPixel DWORDPixel(DWORD c)
{
	FPixel p;
	p.red=(c&0xFF0000)>>8;
	p.green=(c&0xFF00)>>0;
	p.blue=(c&0xFF)<<8;
	return p;
}

void PutPixel(int x, int y, DWORD color)
{
	if(x<0 || y<0 || x>=WINWIDTH || y>=WINHEIGHT)
		return;
	screen[y*WINWIDTH+x]=color;
}


// --

struct Thread
{
	Vector cam_origin;
	Vector cam_x, cam_y, cam_z;
	float cam_pan, cam_tilt, cam_roll;
	
	double anim_t;
	
	bool* ebuffer;
	int* cbuffer;
	float* dbuffer;
	DWORD* debugbuffer;
	
	bool complete;
	bool alive;

	DWORD palette[256];
};

Thread threads[2];

/*
int evals;
int points;

int rpoints[9];
float avrpoints[9];
*/
inline FPixel rgb_blend(FPixel& c1, FPixel& c2, int f)
{
	FPixel r;
	r.red=((c1.red<<8)+(c2.red-c1.red)*f)>>8;
	r.green=((c1.green<<8)+(c2.green-c1.green)*f)>>8;
	r.blue=((c1.blue<<8)+(c2.blue-c1.blue)*f)>>8;
	return r;
}

bool Inside(Vector p, float ft)
{
/*
	p.x+=sin(p.y*5.0f+ft*0.1f)*0.15f;
	p.y+=sin(p.x*4.4f+ft*0.23f)*0.2f;
	if(p.Length()*p.y*p.x-p.z<0.01f)
		return true;
*/

//	evals++;
//	p.x=floor(p.x*4)/4;
//	p.y=floor(p.y*4)/4;
//	p.z=floor(p.z*4)/4;
//	p.x+=cos(p.y*p.z*4.0f);
	p.y*=sin(p.x*2.0f-p.z*1.6f+ft*0.1f)*0.15f;
//	p.y+=sin(p.x*4.4f+ft*0.23f)*0.2f;
	p.x=floor(p.y*4)/4;
//	p.y=floor(p.y*4)/4;
//	if(p.Length()*p.y-p.z<0.02f)
	if(p.Length()*p.y<-0.28f)
		return true;
	return false;
}

float ShootRay(Vector dir, Vector origin, float ft)
{
	float z=0.0f;
	float dz=0.8f;
	float zfar=28.5f;
	bool has_hit=false;
	while(dz>0.0005f && z<zfar)
	{
		if(has_hit)
			dz*=0.5f;
		z+=dz;
		if(Inside(origin+dir*z, ft))
		{
			has_hit=true;
			z-=dz;
		}
	}
	if(z>=zfar)
		z=1000.0f;
	return z;
}
/*
float ProbeRay(Vector dir, Vector origin, float z, bool& probed)
{
	float dz=0.015f;
	if(Inside(origin+dir*(z-dz)))
		return ShootRay(dir, origin);
	if(!Inside(origin+dir*(z+dz)))
		return ShootRay(dir, origin);
	if(Inside(origin+dir*(z-0.1f)))
		return ShootRay(dir, origin);
	float zfar=8.5f;
	bool has_hit=false;
	while(dz>0.0005f && z<zfar)
	{
		if(has_hit)
			dz*=0.5f;
		z+=dz;
		if(Inside(origin+dir*z))
		{
			has_hit=true;
			z-=dz;
		}
	}
	if(z>=zfar)
		z=1000.0f;
	probed=true;
	return z;
}
*/
void Render(int x1, int y1, int x2, int y2, int step, Thread& thread, float ft, bool everything=false)
{
	int stepmask=step*2-1;
//	int yi=0;
	for(int y=y1;y<y2;y+=step)
	{
		float fy=((float)(WINHEIGHT/2-y))/(WINHEIGHT/2);
		float z=0.0f, pz=0.0f, ppz=0.0f;
//		int xi=0;
		for(int x=x1;x<x2;x+=step)
		{
			if(!everything)
			{
				if(step<8 && (x&stepmask)==0 && (y&stepmask)==0)
					continue;
				bool skip=true;
				if(thread.ebuffer[y*BUFFERWIDTH+x]) skip=false;
				int px=x-step;
				if(px>0 && thread.ebuffer[y*BUFFERWIDTH+px]) skip=false;
				int py=y-step;
				if(py>0 && thread.ebuffer[py*BUFFERWIDTH+x]) skip=false;
	//			if(x>WINWIDTH*11/16 && y<WINHEIGHT*1/4) skip=true;
	//			if(x<WINWIDTH*6/16 && y>WINHEIGHT*11/16) skip=true;
				if(skip)
				{
	//				xi=0;
					continue;
				}
/*				if(step<8 && (x&stepmask)==0 && (y&stepmask)==0)
				{
					debugbuffer[y*WINWIDTH+x]=0x00FF00;
					continue;
				}*/
			}
			float fx=((float)(x-WINWIDTH/2))/(WINWIDTH/2)*(4.0f/3.0f);
			Vector raydir=thread.cam_x*fx+thread.cam_y*fy+thread.cam_z*0.8f;
			raydir.Normalize();

/*			if(xi>0 && yi>0)
			{
				float pz=(dbuffer[y*WINWIDTH+px]+dbuffer[py*WINWIDTH+x])/2;
				dbuffer[y*WINWIDTH+x]=ProbeRay(raydir, cam_origin, pz);
			}
			else if(xi>1 && yi>1)
			{
				float pz=(dbuffer[y*WINWIDTH+px]+dbuffer[py*WINWIDTH+x])/2;
				dbuffer[y*WINWIDTH+x]=ProbeRay(raydir, cam_origin, pz);
			}*/
/*			bool probed=false;
			if(xi>1)
			{
				z=ProbeRay(raydir, cam_origin, pz+(pz-ppz), probed);
				if(probed) debugbuffer[y*WINWIDTH+x]=0xFF0000;
			}
			else if(xi>0)
			{
				z=ProbeRay(raydir, cam_origin, pz, probed);
				if(probed) debugbuffer[y*WINWIDTH+x]=0xFF0000;
			}
			else
			{
				z=ShootRay(raydir, cam_origin);
				debugbuffer[y*WINWIDTH+x]=0x000000;
			}*/
			z=ShootRay(raydir, thread.cam_origin, ft);
			if(step==8)
				thread.debugbuffer[y*BUFFERWIDTH+x]=0x0000FF;
			else if(step==4)
				thread.debugbuffer[y*BUFFERWIDTH+x]=0x9955FF;
			else if(step==2)
				thread.debugbuffer[y*BUFFERWIDTH+x]=0xFF77FF;
			else if(step==1)
				thread.debugbuffer[y*BUFFERWIDTH+x]=0xFFFFFF;
			thread.dbuffer[y*BUFFERWIDTH+x]=z;
//			ppz=pz;
//			pz=z;
//			if(fabs(pz-ppz)>0.005f)
//				ppz=pz;
//			xi++;
//			points++;
//			rpoints[step]++;
		}
//		yi++;
/*		if(y%30==0)
		{
			for(int i=0;i<y;i++)
				screen[i*WINWIDTH+4]=0xFFFF00;
			UpdateWindow();
		}*/
	}
	float scale=50.0f/160*WINWIDTH/step;
	for(int y=y1;y<y2;y+=step)
	{
		for(int x=x1;x<x2;x+=step)
		{
			if(!thread.ebuffer[y*BUFFERWIDTH+x])
				continue;
			float r=thread.dbuffer[y*BUFFERWIDTH+x];
			int ic;
			Vector normal=Vector(0,0,0);
			float d, dl;
			int nx=x+step; if(nx>=BUFFERWIDTH) nx=x;
			d=(thread.dbuffer[y*BUFFERWIDTH+nx]-r)*scale;
			dl=sqrt(1.0f+d*d);
//			normal.x+=1.0f/dl;
			normal.z+=1.0f/dl;
			int ny=y+step; if(ny>=BUFFERHEIGHT) ny=y;
			d=(thread.dbuffer[ny*BUFFERWIDTH+x]-r)*scale;
			dl=sqrt(1.0f+d*d);
//			normal.y+=1.0f/sqrt(1.0f+d*d);
			normal.z+=1.0f/dl;
			normal.z*=0.5f;
			ic=(int)(normal.z*255.0f);
			if(ic>255) ic=255;
			if(ic<0) ic=0;
			ic=255-ic;
			if(step==1)
				thread.cbuffer[y*BUFFERWIDTH+x]=ic;
			else
			{
//				int* dp=&cbuffer[y*WINWIDTH+x];
				float* dpd=&thread.dbuffer[y*BUFFERWIDTH+x];
				for(int by=0;by<step;by++)
				{
					for(int bx=0;bx<step;bx++)
					{
//						*dp++=ic;
						*dpd++=r;
					}
//					dp+=WINWIDTH-step;
					dpd+=BUFFERWIDTH-step;
				}
				for(int by=0;by<step;by++)
					thread.cbuffer[(y+by)*BUFFERWIDTH+x]=ic;
				for(int bx=0;bx<step;bx++)
					thread.cbuffer[y*BUFFERWIDTH+x+bx]=ic;
				if(x>=step && y>=step)
				{
/*					for(int by=0;by<step;by++)
						cbuffer[(y+by)*WINWIDTH+x]=ic;
					for(int bx=0;bx<step;bx++)
						cbuffer[y*WINWIDTH+x+bx]=ic;*/
					int c0x=thread.cbuffer[(y-step)*BUFFERWIDTH+x-step];
					int c1x=thread.cbuffer[(y-step)*BUFFERWIDTH+x];
					int c2x=thread.cbuffer[y*BUFFERWIDTH+x-step];
					int c3x=thread.cbuffer[y*BUFFERWIDTH+x];
					int dx0=c1x-c0x;
					int dx1=c3x-c2x;
					int lgstep=3;
					if(step==4) lgstep=2;
					if(step==2) lgstep=1;
					int cxa;
					int cxb;
					int c;
					for(int by=0;by<step;by++)
					{
						int* dp=&thread.cbuffer[(y-step+by)*BUFFERWIDTH+x-step];
						int dx0i=0;
						int dx1i=0;
						for(int bx=0;bx<step;bx++)
						{
							cxa=c0x+(dx0i>>lgstep);
							cxb=c2x+(dx1i>>lgstep);
							dx0i+=dx0;
							dx1i+=dx1;
							c=cxa+(((cxb-cxa)*by)>>lgstep);
							*dp++=c;
						}
					}
				}
//				else
/*				{
					int* dp=&cbuffer[y*WINWIDTH+x];
					for(int by=0;by<step;by++)
					{
						for(int bx=0;bx<step;bx++)
							*dp++=ic;
						dp+=WINWIDTH-step;
					}
				}*/
			}
		}
	}
}

void FillContrastMap(int step, bool value, Thread& thread)
{
	for(int y=0;y<BUFFERHEIGHT;y+=step)
		for(int x=0;x<BUFFERWIDTH;x+=step)
			thread.ebuffer[y*BUFFERWIDTH+x]=value;
}

void BuildContrastMap(int step, int limit, Thread& thread)
{
	for(int y=0;y<BUFFERHEIGHT;y+=step)
		for(int x=0;x<BUFFERWIDTH;x+=step)
			thread.ebuffer[y*BUFFERWIDTH+x]=false;

	for(int y=0;y<BUFFERHEIGHT;y+=step)
		for(int x=0;x<BUFFERWIDTH;x+=step)
		{
			int nx=x+step; if(nx>=BUFFERWIDTH) nx=x;
			int ny=y+step; if(ny>=BUFFERHEIGHT) ny=y;
			int r=thread.cbuffer[y*BUFFERWIDTH+x];
			bool edge=false;
			if(abs(thread.cbuffer[y*BUFFERWIDTH+nx]-r)>limit ||
			   abs(thread.cbuffer[ny*BUFFERWIDTH+x]-r)>limit)
				edge=true;
			if(edge)
			{
				thread.ebuffer[y*BUFFERWIDTH+x]=true;
				thread.ebuffer[y*BUFFERWIDTH+nx]=true;
				thread.ebuffer[ny*BUFFERWIDTH+nx]=true;
				nx=x-step; if(nx>0) thread.ebuffer[y*BUFFERWIDTH+nx]=true;
				ny=y-step; if(ny>0) thread.ebuffer[ny*BUFFERWIDTH+nx]=true;
			}
		}
	for(int y=0;y<BUFFERHEIGHT;y+=step)
		for(int x=0;x<BUFFERWIDTH;x+=step)
		{
			int nx=x+step/2; if(nx>=BUFFERWIDTH) nx=x;
			int ny=y+step/2; if(ny>=WINHEIGHT) ny=y;
			bool r=thread.ebuffer[y*BUFFERWIDTH+x];
			thread.ebuffer[y*BUFFERWIDTH+nx]=r;
			thread.ebuffer[ny*BUFFERWIDTH+x]=r;
			thread.ebuffer[ny*BUFFERWIDTH+nx]=r;
		}
}

void Visualize(Thread& thread)
{
	int i=0;
	for(int y=0;y<WINHEIGHT;y++)
		for(int x=0;x<WINWIDTH;x++)
			screen[i++]=thread.palette[thread.cbuffer[y*BUFFERWIDTH+x]&255];
}

void RotateVectorX(Vector& v, float angle)
{
	Vector v0=v;
	v.y=v0.y*cos(angle)+v0.z*sin(angle);
	v.z=v0.z*cos(angle)-v0.y*sin(angle);
}

void RotateVectorY(Vector& v, float angle)
{
	Vector v0=v;
	v.x=v0.x*cos(angle)-v0.z*sin(angle);
	v.z=v0.z*cos(angle)+v0.x*sin(angle);
}

void RotateVectorZ(Vector& v, float angle)
{
	Vector v0=v;
	v.x=v0.x*cos(angle)-v0.y*sin(angle);
	v.y=v0.y*cos(angle)+v0.x*sin(angle);
}

void MoveCamera(Thread& thread)
{
	thread.cam_x=Vector(1,0,0);
	thread.cam_y=Vector(0,1,0);
	thread.cam_z=Vector(0,0,1);
	RotateVectorZ(thread.cam_x, thread.cam_roll);
	RotateVectorZ(thread.cam_y, thread.cam_roll);
	RotateVectorZ(thread.cam_z, thread.cam_roll);
	RotateVectorX(thread.cam_x, thread.cam_tilt);
	RotateVectorX(thread.cam_y, thread.cam_tilt);
	RotateVectorX(thread.cam_z, thread.cam_tilt);
	RotateVectorY(thread.cam_x, thread.cam_pan);
	RotateVectorY(thread.cam_y, thread.cam_pan);
	RotateVectorY(thread.cam_z, thread.cam_pan);
}

double anim_rate=1.0;
int num_threads;
const float rate_multiplier=0.45f;
//int anim_start=(int)(300/rate_multiplier);
int anim_start=(int)(420/rate_multiplier);
double global_anim_t;

void ProduceFrame(Thread& thread)
{
	float ft=(float)thread.anim_t*rate_multiplier;
	
	float diag=sin(ft*0.01f);
	thread.cam_pan=0.0f+diag*0.6f+sin(ft*0.015f)*0.4f;
	thread.cam_tilt=0.0f-diag*0.6f;
	thread.cam_roll=ft*0.002f;
	MoveCamera(thread);
//	thread.cam_origin=Vector(0,0,0)-thread.cam_z*(1.7f+sin(ft*0.06f)*0.5f);
	thread.cam_origin=Vector(0,0,0)-thread.cam_z*(2.7f+sin(ft*0.06f)*0.5f);

	for(int i=0;i<BUFFERWIDTH*BUFFERHEIGHT;i++)
		thread.debugbuffer[i]=0x000000;

	for(int i=0;i<BUFFERWIDTH*BUFFERHEIGHT;i++)
		thread.cbuffer[i]=0x00;

	if(KeyPressed(Key::B))
	{
		FillContrastMap(1, true, thread);
		Render(0, 0, BUFFERWIDTH, BUFFERHEIGHT, 1, thread, ft, true);
	}
	else
	{
		FillContrastMap(8, true, thread);
		Render(0, 0, BUFFERWIDTH, BUFFERHEIGHT, 8, thread, ft);
		BuildContrastMap(8, 25, thread);
		Render(0, 0, BUFFERWIDTH, BUFFERHEIGHT, 4, thread, ft);
		BuildContrastMap(4, 65, thread);
		Render(0, 0, BUFFERWIDTH, BUFFERHEIGHT, 2, thread, ft);
/*		BuildContrastMap(2, 90, thread);
	//	BuildContrastMap(2, 120, thread);
		Render(0, 0, BUFFERWIDTH-16, BUFFERHEIGHT-16, 1, thread, ft);*/
	}	
//	thread.anim_t+=anim_rate*num_threads;
//	thread.anim_t+=num_threads;
}

void InitThread(Thread& thread, int thread_id)
{
	thread.alive=true;
	thread.complete=true;
//	thread.anim_t=anim_start+thread_id;

	thread.ebuffer=new bool[BUFFERWIDTH*BUFFERHEIGHT];
	thread.cbuffer=new int[BUFFERWIDTH*BUFFERHEIGHT];
	thread.dbuffer=new float[BUFFERWIDTH*BUFFERHEIGHT];
	thread.debugbuffer=new DWORD[BUFFERWIDTH*BUFFERHEIGHT];

	thread.cam_pan=0.0f;
	thread.cam_tilt=0.0f;
	thread.cam_roll=0.0f;
	thread.cam_origin=Vector(0,0,0);

	for(int y=0;y<BUFFERHEIGHT;y++)
		for(int x=0;x<BUFFERWIDTH;x++)
		{
			thread.ebuffer[y*BUFFERWIDTH+x]=false;
			thread.cbuffer[y*BUFFERWIDTH+x]=0;
			thread.dbuffer[y*BUFFERWIDTH+x]=1000.0f;
		}

	for(int i=0;i<256;i++)
	{
		float f=(float)i/255;
		float r=f*2.0f;
		float g=f*1.3f-0.3f;
		float b=f*1.6f-0.9f;
		int ir=(int)(r*255); if(ir<0) ir=0; if(ir>255) ir=255;
		int ig=(int)(g*255); if(ig<0) ig=0; if(ig>255) ig=255;
		int ib=(int)(b*255); if(ib<0) ib=0; if(ib>255) ib=255;
		thread.palette[i]=(ir<<16)|(ig<<8)|ib;
	}
}

void FreeThread(Thread& thread)
{
	delete[] thread.ebuffer;
	delete[] thread.cbuffer;
	delete[] thread.dbuffer;
	delete[] thread.debugbuffer;
}

DWORD WINAPI ThreadFunc(LPVOID lpParam)
{
	int thread_id=(int)lpParam;
	Thread& thread=threads[thread_id];
	while(thread.alive)
	{
		if(thread.complete)
		{
			Sleep(0);
			continue;
		}
//		thread.anim_t=global_anim_t+thread_id;

		ProduceFrame(thread);

		thread.complete=true;
	}
	return 0;
}

void main()
{
	screen=OpenWindow("minild implicit surface action (B for bruteforce, Space for grid)", WINWIDTH, WINHEIGHT);
	if(screen==NULL)
		return;
	
	srand(time(NULL));

//	backbuffers[0]=new FPixel[640*480];
//	metascreen=new MetaPixel[160*120];

	// dispatch two threads for rendering		
	DWORD dwThreadId[16];
	HANDLE hThread[16];
	num_threads=1;
	printf("starting %i threads...\n", num_threads);
	for(int i=0;i<num_threads;i++)
	{
		InitThread(threads[i], i);
		hThread[i]=CreateThread(NULL,  			   // default security attributes
								0,                 // use default stack size  
								(LPTHREAD_START_ROUTINE)ThreadFunc,        // thread function 
								(LPVOID)i,             // argument to thread function 
								0,                 // use default creation flags 
								(LPDWORD)&dwThreadId[i]);   // returns the thread identifier
	}
	
	float evpp=0.0f;
	float avtp=0.0f;

	unsigned int t0=time(NULL);
	int frames=0;
	float cur_fps=25.0f;

	ClearScreen(0x000000);
	
	global_anim_t=anim_start;

	bool seed=false;
//	int it=30;
	for(;;)
	{
		if(KeyPressed(Key::Q) || WindowClosed())
			break;
		ForgetKeys();

/*		anim_t++;		
		it++;
		float t=(float)it*(2.0f/60);
		cam_pan=t;
		cam_tilt=-0.9f+cos(t*1.5f)*0.4f;
		MoveCamera();
		cam_origin=Vector(0,20,15)-cam_z*60.0f;
*/


//		Visualize(thread_id);

/*		if(KeyPressed(Key::Space))
		{
			for(int y=0;y<WINHEIGHT;y+=1)
				for(int x=0;x<WINWIDTH;x+=1)
				{
					int i=y*WINWIDTH+x;
					if(debugbuffer[i]!=0x000000)
						screen[i]=debugbuffer[i];
				}
		}
*/
//		evpp+=((float)evals/points-evpp)*0.2f;
/*		evpp+=((float)evals/(WINWIDTH*WINHEIGHT)-evpp)*0.2f;
		avtp+=((float)points-avtp)*0.2f;

		for(int i=0;i<9;i++)
			avrpoints[i]+=((float)rpoints[i]-avrpoints[i])*0.2f;
*/
//		if(KeyPressed(Key::R) || !seed)

//		for(int i=0;i<640*480;i++)
//			screen[i]=PixelDWORD(backbuffer[i]);

//		ProduceFrame(threads[0]);
//		threads[0].complete=true;

/*		bool done=false;
		while(!done)
		{
			Sleep(0);
			done=true;
			for(int i=0;i<num_threads;i++)
				if(!threads[i].complete)
					done=false;
		}*/

		double next_anim_t=global_anim_t+anim_rate*num_threads;
		
		for(int i=0;i<num_threads;i++)
		{
			while(!threads[i].complete)
				Sleep(0);
			ClearScreen(0x000000);
			Visualize(threads[i]);
			if(KeyPressed(Key::Space))
			{
				int si=0;
				for(int y=0;y<WINHEIGHT;y++)
					for(int x=0;x<WINWIDTH;x++)
					{
						int di=y*BUFFERWIDTH+x;
						if(threads[i].debugbuffer[di]!=0x000000)
							screen[si]=threads[i].debugbuffer[di];
						si++;
					}
			}
			UpdateWindow();
			threads[i].anim_t=next_anim_t+anim_rate*i;
			threads[i].complete=false;
			frames++;
			if(num_threads>1)
			{
				Sleep((int)(800.0f/cur_fps)); // delay between showing frames to keep steady framerate
				if(KeyPressed(Key::B))
					Sleep(400);
			}
		}
		global_anim_t=next_anim_t;
		
//		anim_t++;
//		anim_f=(float)anim_t*0.5f;
//		anim_f=0.0f;
		unsigned int ttime=time(NULL)-t0;
		float new_fps=cur_fps;
		if(ttime>3)
			new_fps=(float)frames/ttime;
		if(new_fps>cur_fps)
			cur_fps=new_fps;
		anim_rate=25.0f/cur_fps;
	}

	unsigned int ttime=time(NULL)-t0;
	printf("%.2f fps\n", (float)frames/ttime);

	printf("closing threads...\n");
	for(int i=0;i<num_threads;i++)	
		threads[i].alive=false;
	WaitForMultipleObjects(num_threads, hThread, TRUE, INFINITE);
	for(int i=0;i<num_threads;i++)
		CloseHandle(hThread[i]);

/*	printf("evals/pixel: %.2f\n", evpp);
	printf("8: %.2f%%\n", avrpoints[8]*100/avtp);
	printf("4: %.2f%%\n", avrpoints[4]*100/avtp);
	printf("2: %.2f%%\n", avrpoints[2]*100/avtp);
	printf("1: %.2f%%\n", avrpoints[1]*100/avtp);
*/
	HKFree();
	CloseWindow();
	
	for(int i=0;i<num_threads;i++)
		FreeThread(threads[i]);
}

